import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)


def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1], padding='SAME')


def conv_net(x, weights, biases):
    conv1 = conv2d(x, weights['wc1'], biases['b1'])
    conv2 = conv2d(conv1, weights['wc2'], biases['b2'])
    conv3 = conv2d(conv2, weights['wc3'], biases['b3'])
    conv4 = conv2d(conv3, weights['wc4'], biases['b4'])

    out = conv4
    return out


def wts_and_bias():
    weights = {'wc1': tf.get_variable('W0', shape=(3, 3, 1, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc2': tf.get_variable('W1', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc3': tf.get_variable('W2', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc4': tf.get_variable('W3', shape=(3, 3, 32, 1), initializer=tf.contrib.layers.xavier_initializer())}

    biases = {'b1': tf.get_variable('B0', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b2': tf.get_variable('B1', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b3': tf.get_variable('B2', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b4': tf.get_variable('B3', shape=(1), initializer=tf.contrib.layers.xavier_initializer())}

    return weights, biases


import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1],padding='SAME')

def conv2d_transpose(x, W, b, output, strides=2):
    # Conv2D transpose wrapper, with bias and relu activation
    x = tf.nn.conv2d_transpose(x, W, output_shape= output, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def make_wts_and_bias(weights, biases, input_channels, output_channels, type):
    curr_weights_num = len(weights)
    curr_biases_num = len(biases)
    if type == 'normal':
        new_w = 'wc' + str(curr_weights_num + 1)
        weights[new_w] = tf.get_variable('W' + str(curr_weights_num + 1), shape=(3, 3, input_channels, output_channels), initializer=tf.contrib.layers.xavier_initializer())
        new_b = 'b' + str(curr_biases_num + 1)
        biases[new_b] = tf.get_variable('B' + str(curr_biases_num + 1), shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    elif type == 'transpose':
        new_w = 'wuc' + str(curr_weights_num + 1)
        weights[new_w] = tf.get_variable('W' + str(curr_weights_num + 1), shape=(3, 3, output_channels, input_channels), initializer=tf.contrib.layers.xavier_initializer())
        new_b = 'ub' + str(curr_biases_num + 1)
        biases[new_b] = tf.get_variable('B' + str(curr_biases_num + 1), shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    return weights, biases, new_w, new_b

def conv_block(x, weights, biases, output_channels):
    input_channels =  x.shape[3]
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, input_channels, output_channels, 'normal')
    y1 = conv2d(x, weights[new_w], biases[new_b])

    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, output_channels, output_channels, 'normal')
    y2 = conv2d(y1, weights[new_w], biases[new_b])

    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, output_channels, output_channels, 'normal')
    y3 = conv2d(y2, weights[new_w], biases[new_b])
    return y1, y2, y3, weights, biases


def conv_net(x):
    weights = {}
    biases = {}

    ####### Unet 1 #########
    # block1
    conv1, conv2, conv3, weights, biases = conv_block(x, weights, biases, 32)

    # block2
    mp1 = maxpool2d(conv3, 2)
    conv4, conv5, conv6, weights, biases = conv_block(mp1, weights, biases, 32)

    # block3
    mp2 = maxpool2d(conv6, 2)
    conv7, conv8, conv9, weights, biases = conv_block(mp2, weights, biases, 32)

    # block4
    mp3 = maxpool2d(conv9, 2)
    conv10, conv11, conv12, weights, biases = conv_block(mp3, weights, biases, 32)

    # block5
    mp4 = maxpool2d(conv12, 2)
    conv13, conv14, conv15, weights, biases = conv_block(mp4, weights, biases, 32)

    # block6
    mp5 = maxpool2d(conv15, 2)
    conv16, conv17, conv18, weights, biases = conv_block(mp5, weights, biases, 32)

    # up block 1
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv1 = conv2d_transpose(conv18, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 54, 41, 32], strides=2)
    uconv1 = tf.concat([uconv1, conv15], axis=3)
    conv19, conv20, conv21, weights, biases = conv_block(uconv1, weights, biases, 32)

    # up block 2
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv2 = conv2d_transpose(conv21, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 108, 82, 32], strides=2)
    uconv2 = tf.concat([uconv2, conv12], axis=3)
    conv22, conv23, conv24, weights, biases = conv_block(uconv2, weights, biases, 32)

    # up block 3
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv3 = conv2d_transpose(conv24, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 215, 163, 32], strides=2)
    uconv3 = tf.concat([uconv3, conv9], axis=3)
    conv25, conv26, conv27, weights, biases = conv_block(uconv3, weights, biases, 32)

    # up block 4
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv4 = conv2d_transpose(conv27, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 430, 325, 32], strides=2)
    uconv4 = tf.concat([uconv4, conv6], axis=3)
    conv28, conv29, conv30, weights, biases = conv_block(uconv4, weights, biases, 32)

    # up block 5
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv5 = conv2d_transpose(conv30, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 860, 650, 32], strides=2)
    uconv5 = tf.concat([uconv5, conv3], axis=3)
    conv31, conv32, conv33, weights, biases = conv_block(uconv5, weights, biases, 32)

    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 1, 'normal')
    conv34 = conv2d(conv33, weights[new_w], biases[new_b])
    out1 = conv34

    ####### Unet 2 #######
    conv33 = tf.concat([conv33, x], axis=3)

    # block1
    conv35, conv36, conv37, weights, biases = conv_block(conv33, weights, biases, 32)

    # block2
    mp6 = maxpool2d(conv37, 2)
    conv38, conv39, conv40, weights, biases = conv_block(mp6, weights, biases, 32)

    # block3
    mp7 = maxpool2d(conv40, 2)
    conv41, conv42, conv43, weights, biases = conv_block(mp7, weights, biases, 32)

    # block4
    mp8 = maxpool2d(conv43, 2)
    conv44, conv45, conv46, weights, biases = conv_block(mp8, weights, biases, 32)

    # block5
    mp9 = maxpool2d(conv46, 2)
    conv47, conv48, conv49, weights, biases = conv_block(mp9, weights, biases, 32)

    # block6
    mp10 = maxpool2d(conv49, 2)
    conv50, conv51, conv52, weights, biases = conv_block(mp10, weights, biases, 32)

    # up block 1
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv6 = conv2d_transpose(conv52, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 54, 41, 32], strides=2)
    uconv6 = tf.concat([uconv6, conv49], axis=3)
    conv53, conv54, conv55, weights, biases = conv_block(uconv6, weights, biases, 32)

    # up block 2
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv7 = conv2d_transpose(conv55, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 108, 82, 32], strides=2)
    uconv7 = tf.concat([uconv7, conv46], axis=3)
    conv56, conv57, conv58, weights, biases = conv_block(uconv7, weights, biases, 32)

    # up block 3
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv8 = conv2d_transpose(conv58, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 215, 163, 32], strides=2)
    uconv8 = tf.concat([uconv8, conv43], axis=3)
    conv59, conv60, conv61, weights, biases = conv_block(uconv8, weights, biases, 32)

    # up block 4
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv9 = conv2d_transpose(conv61, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 430, 325, 32], strides=2)
    uconv9 = tf.concat([uconv9, conv40], axis=3)
    conv62, conv63, conv64, weights, biases = conv_block(uconv9, weights, biases, 32)

    # up block 5
    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 32, 'transpose')
    uconv10 = conv2d_transpose(conv64, weights[new_w], biases[new_b], [tf.shape(conv3)[0], 860, 650, 32], strides=2)
    uconv10 = tf.concat([uconv10, conv37], axis=3)
    conv65, conv66, conv67, weights, biases = conv_block(uconv10, weights, biases, 32)

    weights, biases, new_w, new_b = make_wts_and_bias(weights, biases, 32, 1, 'normal')
    conv68 = conv2d(conv67, weights[new_w], biases[new_b])
    out2 = conv68

    return out1, out2


import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1],padding='SAME')

def conv2d_transpose(x, W, b, output, strides=2):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d_transpose(x, W, output_shape= output, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)


def conv_net(x, weights, biases):
    # block1
    conv1 = conv2d(x, weights['wc1'], biases['b1'])
    conv2 = conv2d(conv1, weights['wc2'], biases['b2'])
    conv3 = conv2d(conv2, weights['wc3'], biases['b3'])

    # block2
    mp1 = maxpool2d(conv3, 2)
    conv4 = conv2d(mp1, weights['wc4'], biases['b4'])
    conv5 = conv2d(conv4, weights['wc5'], biases['b5'])
    conv6 = conv2d(conv5, weights['wc6'], biases['b6'])

    # block3
    mp2 = maxpool2d(conv6, 2)
    conv7 = conv2d(mp2, weights['wc7'], biases['b7'])
    conv8 = conv2d(conv7, weights['wc8'], biases['b8'])
    conv9 = conv2d(conv8, weights['wc9'], biases['b9'])

    # up block 1
    uconv2 = conv2d_transpose(conv9, weights['wuc2'], biases['ub2'], [tf.shape(conv3)[0], 430, 325, 64], strides=2)
    conv16 = conv2d(uconv2, weights['wc16'], biases['b16'])
    conv17 = conv2d(conv16, weights['wc17'], biases['b17'])
    conv18 = conv2d(conv17, weights['wc18'], biases['b18'])

    # up block 2
    uconv3 = conv2d_transpose(conv18, weights['wuc3'], biases['ub3'], [tf.shape(conv3)[0], 860, 650, 32], strides=2)
    conv19 = conv2d(uconv3, weights['wc19'], biases['b19'])
    conv20 = conv2d(conv19, weights['wc20'], biases['b20'])
    conv21 = conv2d(conv20, weights['wc21'], biases['b21'])

    conv22 = conv2d(conv21, weights['wc22'], biases['b22'])
    out = conv22
    return out


def wts_and_bias():
    weights = {'wc1': tf.get_variable('W0', shape=(3, 3, 1, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc2': tf.get_variable('W1', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc3': tf.get_variable('W2', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc4': tf.get_variable('W3', shape=(3, 3, 32, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc5': tf.get_variable('W4', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc6': tf.get_variable('W5', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc7': tf.get_variable('W6', shape=(3, 3, 64, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc8': tf.get_variable('W7', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc9': tf.get_variable('W8', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc2': tf.get_variable('W16', shape=(3, 3, 64, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc16': tf.get_variable('W17', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc17': tf.get_variable('W18', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc18': tf.get_variable('W19', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc3': tf.get_variable('W20', shape=(3, 3, 32, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc19': tf.get_variable('W21', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc20': tf.get_variable('W22', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc21': tf.get_variable('W23', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc22': tf.get_variable('W24', shape=(1, 1, 32, 1), initializer=tf.contrib.layers.xavier_initializer())}

    biases = {'b1': tf.get_variable('B0', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b2': tf.get_variable('B1', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b3': tf.get_variable('B2', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b4': tf.get_variable('B3', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b5': tf.get_variable('B4', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b6': tf.get_variable('B5', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b7': tf.get_variable('B6', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b8': tf.get_variable('B7', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b9': tf.get_variable('B8', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'ub2': tf.get_variable('B16', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b16': tf.get_variable('B17', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b17': tf.get_variable('B18', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b18': tf.get_variable('B19', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'ub3': tf.get_variable('B20', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b19': tf.get_variable('B21', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b20': tf.get_variable('B22', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b21': tf.get_variable('B23', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b22': tf.get_variable('B24', shape=(1), initializer=tf.contrib.layers.xavier_initializer()),}

    return weights, biases


import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1],padding='SAME')

def conv2d_transpose(x, W, b, output, strides=2):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d_transpose(x, W, output_shape= output, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def make_wts_and_bias(var_suffix, input_channels, output_channels, type):
    if type == 'normal':
        weights = tf.get_variable('W' + var_suffix, shape=(3, 3, input_channels, output_channels), initializer=tf.contrib.layers.xavier_initializer())
        biases = tf.get_variable('B' + var_suffix, shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    elif type == 'transpose':
        weights = tf.get_variable('W' + var_suffix, shape=(3, 3, output_channels, input_channels), initializer=tf.contrib.layers.xavier_initializer())
        biases = tf.get_variable('B' + var_suffix, shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    return weights, biases

def residual_block(x, output_channels, scope_name):
    input_channels =  x.shape[3]
    with tf.variable_scope(scope_name, reuse= tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('1', input_channels, output_channels, 'normal')
    y1 = conv2d(x, weights, biases)

    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('2', output_channels, output_channels, 'normal')
    y2 = conv2d(x, weights, biases)

    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('3', output_channels, output_channels, 'normal')
    y3 = conv2d(x, weights, biases)

    y3 = y3 + x
    return y1, y2, y3


def conv_net(x, output_shape):

    # start
    with tf.variable_scope('start', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('1', x.shape[3], 32, 'normal')
    conv0 = conv2d(x, weights, biases)

    # block1
    conv1, conv2, conv3 = residual_block(conv0, 32, 'block1')
    mp1 = maxpool2d(conv3, 2)

    # block2
    conv4, conv5, conv6 = residual_block(mp1, 32, 'block2')
    mp2 = maxpool2d(conv6, 2)

    # block3
    conv7, conv8, conv9 = residual_block(mp2, 32, 'block3')
    mp3 = maxpool2d(conv9, 2)

    # block4
    conv10, conv11, conv12 = residual_block(mp3, 32, 'block4')

    # sideblock1
    conv13, conv14, conv15 = residual_block(conv3, 32, 'sideblock1')

    # sideblock2
    conv16, conv17, conv18 = residual_block(conv6, 32, 'sideblock2')

    # sideblock3
    conv19, conv20, conv21 = residual_block(conv9, 32, 'sideblock3')

    # up1
    mp4 = maxpool2d(conv21, 2)
    conv12 = conv12 + mp4
    with tf.variable_scope('up_block1', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('U1', 32, 32, 'transpose')
    conv22 = conv2d_transpose(conv12, weights, biases, [tf.shape(conv12)[0], 128, 128, 32], strides=2)
    conv23, conv24, conv25 = residual_block(conv22, 32, 'up_block1')

    # up2
    mp5 = maxpool2d(conv18, 2)
    conv25 = conv25 + mp5
    with tf.variable_scope('up_block2', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('U1', 32, 32, 'transpose')
    conv26 = conv2d_transpose(conv25, weights, biases, [tf.shape(conv12)[0], 256, 256, 32], strides=2)
    conv27, conv28, conv29 = residual_block(conv26, 32, 'up_block2')

    # up3
    mp6 = maxpool2d(conv15, 2)
    conv29 = conv29 + mp6
    with tf.variable_scope('up_block3', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('U1', 32, 32, 'transpose')
    conv30 = conv2d_transpose(conv29, weights, biases, [tf.shape(conv12)[0], 512, 512, 32], strides=2)
    conv31, conv32, conv33 = residual_block(conv30, 32, 'up_block3')

    # proj
    with tf.variable_scope('proj', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('1', 32, output_shape, 'normal')
    conv34 = conv2d(conv33, weights, biases)
    out = conv34
    return out


import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1],padding='SAME')

def conv2d_transpose(x, W, b, output, strides=2):
    # Conv2D transpose wrapper, with bias and relu activation
    x = tf.nn.conv2d_transpose(x, W, output_shape= output, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def make_wts_and_bias(var_suffix, input_channels, output_channels, type):
    if type == 'normal':
        weights = tf.get_variable('W' + var_suffix, shape=(3, 3, input_channels, output_channels), initializer=tf.contrib.layers.xavier_initializer())
        biases = tf.get_variable('B' + var_suffix, shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    elif type == 'transpose':
        weights = tf.get_variable('W' + var_suffix, shape=(3, 3, output_channels, input_channels), initializer=tf.contrib.layers.xavier_initializer())
        biases = tf.get_variable('B' + var_suffix, shape=(output_channels), initializer=tf.contrib.layers.xavier_initializer())
    return weights, biases

def conv_block(x, output_channels, scope_name):
    input_channels =  x.shape[3]
    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('1', input_channels, output_channels, 'normal')
    y1 = conv2d(x, weights, biases)

    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('2', output_channels, output_channels, 'normal')
    y2 = conv2d(y1, weights, biases)

    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('3', output_channels, output_channels, 'normal')
    y3 = conv2d(y2, weights, biases)
    return y1, y2, y3


def conv_net(x, output_shape, output_name):

    ####### Unet 1 #########
    # block1
    conv1, conv2, conv3 = conv_block(x, 32, 'block1')

    # block2
    mp1 = maxpool2d(conv3, 2)
    conv4, conv5, conv6 = conv_block(mp1, 32, 'block2')

    # block3
    mp2 = maxpool2d(conv6, 2)
    conv7, conv8, conv9  = conv_block(mp2, 32, 'block3')

    # block4
    mp3 = maxpool2d(conv9, 2)
    conv10, conv11, conv12 = conv_block(mp3, 32, 'block4')

    # block5
    mp4 = maxpool2d(conv12, 2)
    conv13, conv14, conv15 = conv_block(mp4, 32, 'block5')

    # up block 1
    with tf.variable_scope('up_block1', reuse=tf.AUTO_REUSE) as scope:
        up_w, up_b = make_wts_and_bias('U1', 32, 32, 'transpose')
    uconv1 = conv2d_transpose(conv15, up_w, up_b, [tf.shape(conv3)[0], 64, 64, 32], strides=2)
    uconv1 = tf.concat([uconv1, conv12], axis=3)
    conv16, conv17, conv18 = conv_block(uconv1, 32, 'up_block1')

    # up block 2
    with tf.variable_scope('up_block2', reuse=tf.AUTO_REUSE) as scope:
        up_w, up_b = make_wts_and_bias('U1', 32, 32, 'transpose')
    uconv2 = conv2d_transpose(conv18, up_w, up_b, [tf.shape(conv3)[0], 128, 128, 32], strides=2)
    uconv2 = tf.concat([uconv2, conv9], axis=3)
    conv19, conv20, conv21 = conv_block(uconv2, 32, 'up_block2')

    # up block 3
    with tf.variable_scope('up_block3', reuse=tf.AUTO_REUSE) as scope:
        up_w, up_b = make_wts_and_bias('U1', 32, 32, 'transpose')
    uconv3 = conv2d_transpose(conv21, up_w, up_b, [tf.shape(conv3)[0], 256, 256, 32], strides=2)
    uconv3 = tf.concat([uconv3, conv6], axis=3)
    conv22, conv23, conv24 = conv_block(uconv3, 32, 'up_block3')

    # up block 4
    with tf.variable_scope('up_block4', reuse=tf.AUTO_REUSE) as scope:
        up_w, up_b = make_wts_and_bias('U1', 32, 32, 'transpose')
    uconv4 = conv2d_transpose(conv24, up_w, up_b, [tf.shape(conv3)[0], 512, 512, 32], strides=2)
    uconv4 = tf.concat([uconv4, conv3], axis=3)
    conv25, conv26, conv27 = conv_block(uconv4, 32, 'up_block4')

    # proj_to_img
    with tf.variable_scope('proj', reuse=tf.AUTO_REUSE) as scope:
        weights, biases = make_wts_and_bias('p', 32, output_shape, 'normal')
    conv28 = conv2d(conv27, weights, biases)
    out1 = tf.identity(conv28, name= output_name)

    return out1


from setuptools import setup, find_packages

VERSION = '0.1.2'
DESCRIPTION = 'Neuro Imaging Denoising via Deep Learning (NIDDL)'
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
# LONG_DESCRIPTION = 'NIDDL code for denoising of volumetric calcium imaging recordings. \
#         If you find our code useful please cite \
#         Chaudhary, S., Moon, S. & Lu, H. Fast, efficient, and accurate neuro-imaging denoising via supervised deep learning. Nat Commun 13, 5165 (2022). https://doi.org/10.1038/s41467-022-32886-w'

setup(
    name="NIDDL",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Shivesh Chaudhary",
    author_email="shiveshc@gmail.com",
    license='MIT',
    packages=find_packages(),
    url='https://github.com/shiveshc/whole-brain_DeepDenoising',
    install_requires=[
        'absl-py==0.11.0',
        'astor==0.8.1',
        'bleach==1.5.0',
        'certifi==2020.6.20',
        'cycler==0.10.0',
        'Cython==0.29.21',
        'decorator==4.4.2',
        'gast==0.4.0',
        'google-pasta==0.2.0',
        'grpcio==1.35.0',
        'h5py==2.10.0',
        'html5lib==0.9999999',
        'imagecodecs==2019.12.31',
        'imagecodecs-lite==2020.1.31',
        'imageio==2.9.0',
        'importlib-metadata==2.1.1',
        'Keras==2.0.8',
        'Keras-Applications==1.0.8',
        'Keras-Preprocessing==1.1.2',
        'kiwisolver==1.1.0',
        'Markdown==3.2.2',
        'matplotlib==3.0.3',
        'networkx==2.4',
        'numpy==1.16.6',
        'opencv-python==4.4.0.42',
        'Pillow==7.2.0',
        'protobuf==3.14.0',
        'pyparsing==2.4.7',
        'python-dateutil==2.8.1',
        'PyWavelets==1.1.1',
        'PyYAML==5.3.1',
        'scikit-image==0.15.0',
        'scipy==1.4.1',
        'six==1.15.0',
        'tensorflow-gpu==1.6.0',
        'termcolor==1.1.0',
        'tifffile==2019.7.26.2',
        'tqdm==4.61.2',
        'Werkzeug==1.0.1',
        'wrapt==1.12.1',
        'zipp==1.2.0',
    ],
    keywords='conversion',
    classifiers= [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ]
)


import tensorflow as tf

def conv2d(x, W, b, strides=1):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d(x, W, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)

def maxpool2d(x, k=2):
    return tf.nn.max_pool(x, ksize=[1, k, k, 1], strides=[1, k, k, 1],padding='SAME')

def conv2d_transpose(x, W, b, output, strides=2):
    # Conv2D wrapper, with bias and relu activation
    x = tf.nn.conv2d_transpose(x, W, output_shape= output, strides=[1, strides, strides, 1], padding='SAME')
    x = tf.nn.bias_add(x, b)
    return tf.nn.relu(x)


def conv_net(x, weights, biases):
    # block1
    conv1 = conv2d(x, weights['wc1'], biases['b1'])
    conv2 = conv2d(conv1, weights['wc2'], biases['b2'])
    conv3 = conv2d(conv2, weights['wc3'], biases['b3'])

    # block2
    mp1 = maxpool2d(conv3, 2)
    conv4 = conv2d(mp1, weights['wc4'], biases['b4'])
    conv5 = conv2d(conv4, weights['wc5'], biases['b5'])
    conv6 = conv2d(conv5, weights['wc6'], biases['b6'])

    # block3
    mp2 = maxpool2d(conv6, 2)
    conv7 = conv2d(mp2, weights['wc7'], biases['b7'])
    conv8 = conv2d(conv7, weights['wc8'], biases['b8'])
    conv9 = conv2d(conv8, weights['wc9'], biases['b9'])

    # block4
    mp3 = maxpool2d(conv9, 2)
    conv10 = conv2d(mp3, weights['wc10'], biases['b10'])
    conv11 = conv2d(conv10, weights['wc11'], biases['b11'])
    conv12 = conv2d(conv11, weights['wc12'], biases['b12'])

    # block5
    mp4 = maxpool2d(conv12, 2)
    conv13 = conv2d(mp4, weights['wc13'], biases['b13'])
    conv14 = conv2d(conv13, weights['wc14'], biases['b14'])
    conv15 = conv2d(conv14, weights['wc15'], biases['b15'])

    # block6
    mp5 = maxpool2d(conv15, 2)
    conv16 = conv2d(mp5, weights['wc16'], biases['b16'])
    conv17 = conv2d(conv16, weights['wc17'], biases['b17'])
    conv18 = conv2d(conv17, weights['wc18'], biases['b18'])

    # up block 1
    uconv1 = conv2d_transpose(conv18, weights['wuc1'], biases['ub1'], [tf.shape(conv3)[0], 54, 41, 1024], strides=2)
    uconv1 = conv2d(uconv1, weights['wrc1'], biases['rcb1'])
    uconv1 = tf.concat([uconv1, conv15], axis=3)
    conv19 = conv2d(uconv1, weights['wc19'], biases['b19'])
    conv20 = conv2d(conv19, weights['wc20'], biases['b20'])
    conv21 = conv2d(conv20, weights['wc21'], biases['b21'])

    # up block 2
    uconv2 = conv2d_transpose(conv21, weights['wuc2'], biases['ub2'], [tf.shape(conv3)[0], 108, 82, 512], strides=2)
    uconv2 = conv2d(uconv2, weights['wrc2'], biases['rcb2'])
    uconv2 = tf.concat([uconv2, conv12], axis=3)
    conv22 = conv2d(uconv2, weights['wc22'], biases['b22'])
    conv23 = conv2d(conv22, weights['wc23'], biases['b23'])
    conv24 = conv2d(conv23, weights['wc24'], biases['b24'])

    # up block 3
    uconv3 = conv2d_transpose(conv24, weights['wuc3'], biases['ub3'], [tf.shape(conv3)[0], 215, 163, 256], strides=2)
    uconv3 = conv2d(uconv3, weights['wrc3'], biases['rcb3'])
    uconv3 = tf.concat([uconv3, conv9], axis=3)
    conv25 = conv2d(uconv3, weights['wc25'], biases['b25'])
    conv26 = conv2d(conv25, weights['wc26'], biases['b26'])
    conv27 = conv2d(conv26, weights['wc27'], biases['b27'])

    # up block 4
    uconv4 = conv2d_transpose(conv27, weights['wuc4'], biases['ub4'], [tf.shape(conv3)[0], 430, 325, 128], strides=2)
    uconv4 = conv2d(uconv4, weights['wrc4'], biases['rcb4'])
    uconv4 = tf.concat([uconv4, conv6], axis=3)
    conv28 = conv2d(uconv4, weights['wc28'], biases['b28'])
    conv29 = conv2d(conv28, weights['wc29'], biases['b29'])
    conv30 = conv2d(conv29, weights['wc30'], biases['b30'])

    # up block 5
    uconv5 = conv2d_transpose(conv30, weights['wuc5'], biases['ub5'], [tf.shape(conv3)[0], 860, 650, 64], strides=2)
    uconv5 = conv2d(uconv5, weights['wrc5'], biases['rcb5'])
    uconv5 = tf.concat([uconv5, conv3], axis=3)
    conv31 = conv2d(uconv5, weights['wc31'], biases['b31'])
    conv32 = conv2d(conv31, weights['wc32'], biases['b32'])
    conv33 = conv2d(conv32, weights['wc33'], biases['b33'])


    conv34 = conv2d(conv33, weights['wc34'], biases['b34'])
    out = conv34
    return out


def wts_and_bias():
    weights = {'wc1': tf.get_variable('W0', shape=(3, 3, 1, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc2': tf.get_variable('W1', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc3': tf.get_variable('W2', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc4': tf.get_variable('W3', shape=(3, 3, 32, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc5': tf.get_variable('W4', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc6': tf.get_variable('W5', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc7': tf.get_variable('W6', shape=(3, 3, 64, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc8': tf.get_variable('W7', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc9': tf.get_variable('W8', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc10': tf.get_variable('W9', shape=(3, 3, 128, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc11': tf.get_variable('W10', shape=(3, 3, 256, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc12': tf.get_variable('W11', shape=(3, 3, 256, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc13': tf.get_variable('W12', shape=(3, 3, 256, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc14': tf.get_variable('W13', shape=(3, 3, 512, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc15': tf.get_variable('W14', shape=(3, 3, 512, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc16': tf.get_variable('W15', shape=(3, 3, 512, 1024), initializer=tf.contrib.layers.xavier_initializer()),
               'wc17': tf.get_variable('W16', shape=(3, 3, 1024, 1024), initializer=tf.contrib.layers.xavier_initializer()),
               'wc18': tf.get_variable('W17', shape=(3, 3, 1024, 1024), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc1': tf.get_variable('W18', shape=(3, 3, 1024, 1024), initializer=tf.contrib.layers.xavier_initializer()),
               'wrc1': tf.get_variable('W19', shape=(3, 3, 1024, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc19': tf.get_variable('W20', shape=(3, 3, 1024, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc20': tf.get_variable('W21', shape=(3, 3, 512, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wc21': tf.get_variable('W22', shape=(3, 3, 512, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc2': tf.get_variable('W23', shape=(3, 3, 512, 512), initializer=tf.contrib.layers.xavier_initializer()),
               'wrc2': tf.get_variable('W24', shape=(3, 3, 512, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc22': tf.get_variable('W25', shape=(3, 3, 512, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc23': tf.get_variable('W26', shape=(3, 3, 256, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wc24': tf.get_variable('W27', shape=(3, 3, 256, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc3': tf.get_variable('W28', shape=(3, 3, 256, 256), initializer=tf.contrib.layers.xavier_initializer()),
               'wrc3': tf.get_variable('W29', shape=(3, 3, 256, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc25': tf.get_variable('W30', shape=(3, 3, 256, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc26': tf.get_variable('W31', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wc27': tf.get_variable('W32', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc4': tf.get_variable('W33', shape=(3, 3, 128, 128), initializer=tf.contrib.layers.xavier_initializer()),
               'wrc4': tf.get_variable('W34', shape=(3, 3, 128, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc28': tf.get_variable('W35', shape=(3, 3, 128, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc29': tf.get_variable('W36', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wc30': tf.get_variable('W37', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wuc5': tf.get_variable('W38', shape=(3, 3, 64, 64), initializer=tf.contrib.layers.xavier_initializer()),
               'wrc5': tf.get_variable('W39', shape=(3, 3, 64, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc31': tf.get_variable('W40', shape=(3, 3, 64, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc32': tf.get_variable('W41', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc33': tf.get_variable('W42', shape=(3, 3, 32, 32), initializer=tf.contrib.layers.xavier_initializer()),
               'wc34': tf.get_variable('W43', shape=(1, 1, 32, 1), initializer=tf.contrib.layers.xavier_initializer())}

    biases = {'b1': tf.get_variable('B0', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b2': tf.get_variable('B1', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b3': tf.get_variable('B2', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b4': tf.get_variable('B3', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b5': tf.get_variable('B4', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b6': tf.get_variable('B5', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b7': tf.get_variable('B6', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b8': tf.get_variable('B7', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b9': tf.get_variable('B8', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b10': tf.get_variable('B9', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b11': tf.get_variable('B10', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b12': tf.get_variable('B11', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b13': tf.get_variable('B12', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b14': tf.get_variable('B13', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b15': tf.get_variable('B14', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b16': tf.get_variable('B15', shape=(1024), initializer=tf.contrib.layers.xavier_initializer()),
              'b17': tf.get_variable('B16', shape=(1024), initializer=tf.contrib.layers.xavier_initializer()),
              'b18': tf.get_variable('B17', shape=(1024), initializer=tf.contrib.layers.xavier_initializer()),
              'ub1': tf.get_variable('B18', shape=(1024), initializer=tf.contrib.layers.xavier_initializer()),
              'rcb1': tf.get_variable('B19', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b19': tf.get_variable('B20', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b20': tf.get_variable('B21', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'b21': tf.get_variable('B22', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'ub2': tf.get_variable('B23', shape=(512), initializer=tf.contrib.layers.xavier_initializer()),
              'rcb2': tf.get_variable('B24', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b22': tf.get_variable('B25', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b23': tf.get_variable('B26', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'b24': tf.get_variable('B27', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'ub3': tf.get_variable('B28', shape=(256), initializer=tf.contrib.layers.xavier_initializer()),
              'rcb3': tf.get_variable('B29', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b25': tf.get_variable('B30', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b26': tf.get_variable('B31', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'b27': tf.get_variable('B32', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'ub4': tf.get_variable('B33', shape=(128), initializer=tf.contrib.layers.xavier_initializer()),
              'rcb4': tf.get_variable('B34', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b28': tf.get_variable('B35', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b29': tf.get_variable('B36', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'b30': tf.get_variable('B37', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'ub5': tf.get_variable('B38', shape=(64), initializer=tf.contrib.layers.xavier_initializer()),
              'rcb5': tf.get_variable('B39', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b31': tf.get_variable('B40', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b32': tf.get_variable('B41', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b33': tf.get_variable('B42', shape=(32), initializer=tf.contrib.layers.xavier_initializer()),
              'b34': tf.get_variable('B43', shape=(1), initializer=tf.contrib.layers.xavier_initializer()),}

    return weights, biases



__all__ = ['unet_v',
           'unet_v2',
           'hourglass_wres',
           'hourglass_wores',
           'unet_v_synth',
           'unet_v2_synth',
           'hourglass_wres_synth',
           'hourglass_wores_synth',
           'unet_v_tr',
           'hourglass_wres_tr',
           'unet_v_k5',
           'hourglass_wres_k5',
           'unet_v_patch',
           'hourglass_wres_patch']